/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file alsa_writer.c paraslash's alsa output plugin */

/*
 * Based in parts on aplay.c from the alsa-utils-1.0.8 package,
 * Copyright (c) by Jaroslav Kysela <perex@suse.cz>, which is
 * based on the vplay program by Michael Beck.
 */

#include "para.h"
#include "fd.h"
#include "string.h"
#include "write.h"

#include <alsa/asoundlib.h>

#include "write.cmdline.h"
#include "error.h"

extern struct gengetopt_args_info conf;

#define FORMAT SND_PCM_FORMAT_S16_LE

/** data specific to the alsa writer */
struct private_alsa_data {
/** the alsa handle */
snd_pcm_t *handle;
/** determined and set by alsa_open() */
size_t bytes_per_frame;
};

/*
 * open and prepare the PCM handle for writing
 *
 * Install PCM software and hardware configuration. Exit on errors.
 */
static int alsa_open(struct writer_node *w)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t buffer_size, xfer_align, start_threshold,
		stop_threshold;
	unsigned buffer_time = 0;
	int err;
	snd_pcm_info_t *info;
	snd_output_t *log;
	snd_pcm_uframes_t period_size;
	struct private_alsa_data *pad = para_malloc(sizeof(struct private_alsa_data));
	w->private_data = pad;

	snd_pcm_info_alloca(&info);
	if (snd_output_stdio_attach(&log, stderr, 0) < 0)
		return -E_ALSA_LOG;
	err = snd_pcm_open(&pad->handle, conf.device_arg,
		SND_PCM_STREAM_PLAYBACK, 0);
	if (err < 0)
		return -E_PCM_OPEN;
	if ((err = snd_pcm_info(pad->handle, info)) < 0)
		return -E_SND_PCM_INFO;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(pad->handle, hwparams) < 0)
		return -E_BROKEN_CONF;
	if (snd_pcm_hw_params_set_access(pad->handle, hwparams,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		return -E_ACCESS_TYPE;
	if (snd_pcm_hw_params_set_format(pad->handle, hwparams, FORMAT) < 0)
		return -E_SAMPLE_FORMAT;
	if (snd_pcm_hw_params_set_channels(pad->handle, hwparams,
			conf.channels_arg) < 0)
		return -E_CHANNEL_COUNT;
	if (snd_pcm_hw_params_set_rate_near(pad->handle, hwparams,
			(unsigned int*) &conf.sample_rate_arg, 0) < 0)
		return -E_SET_RATE;
	err = snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0);
	if (err < 0 || !buffer_time)
		return -E_GET_BUFFER_TIME;
	PARA_DEBUG_LOG("buffer time: %d\n", buffer_time);
	if (snd_pcm_hw_params_set_buffer_time_near(pad->handle, hwparams,
			&buffer_time, 0) < 0)
		return -E_SET_BUFFER_TIME;
	if (snd_pcm_hw_params(pad->handle, hwparams) < 0)
		return -E_HW_PARAMS;
	snd_pcm_hw_params_get_period_size(hwparams, &period_size, 0);
	snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size);
	PARA_DEBUG_LOG("buffer size: %lu, period_size: %lu\n", buffer_size,
		period_size);
	if (period_size == buffer_size)
		return -E_BAD_PERIOD;
	snd_pcm_sw_params_current(pad->handle, swparams);
	err = snd_pcm_sw_params_get_xfer_align(swparams, &xfer_align);
	if (err < 0 || !xfer_align)
		return -E_GET_XFER;
	snd_pcm_sw_params_set_avail_min(pad->handle, swparams, period_size);
	/* round to closest transfer boundary */
	start_threshold = (buffer_size / xfer_align) * xfer_align;
	if (start_threshold < 1)
		start_threshold = 1;
	if (snd_pcm_sw_params_set_start_threshold(pad->handle, swparams,
			start_threshold) < 0)
		return -E_START_THRESHOLD;
	stop_threshold = buffer_size;
	if (snd_pcm_sw_params_set_stop_threshold(pad->handle, swparams,
			stop_threshold) < 0)
		return -E_STOP_THRESHOLD;
	if (snd_pcm_sw_params_set_xfer_align(pad->handle, swparams,
			xfer_align) < 0)
		return -E_SET_XFER;
	if (snd_pcm_sw_params(pad->handle, swparams) < 0)
		return -E_SW_PARAMS;
	pad->bytes_per_frame = snd_pcm_format_physical_width(FORMAT)
		* conf.channels_arg / 8;
	return period_size * pad->bytes_per_frame;
}

/**
 * push out pcm frames
 * \param data pointer do data to be written
 * \param nbytes number of bytes (not frames)
 *
 * \return Number of bytes written, -E_ALSA_WRITE on errors.
 */
static int alsa_write(char *data, size_t nbytes, struct writer_node *wn)
{
	struct private_alsa_data *pad = wn->private_data;
	size_t frames = nbytes / pad->bytes_per_frame;
	unsigned char *d = (unsigned char*)data;
	snd_pcm_sframes_t r, result = 0;

	while (frames > 0) {
		/* write interleaved frames */
		r = snd_pcm_writei(pad->handle, d, frames);
		if (r < 0)
			PARA_ERROR_LOG("write error: %s\n", snd_strerror(r));
		if (r == -EAGAIN || (r >= 0 && r < frames))
			snd_pcm_wait(pad->handle, 1);
		else if (r == -EPIPE)
			snd_pcm_prepare(pad->handle);
		else if (r < 0)
			return -E_ALSA_WRITE;
		if (r > 0) {
			result += r;
			frames -= r;
			d += r * pad->bytes_per_frame;
		}
	}
	return result * pad->bytes_per_frame;
}

static void alsa_close(struct writer_node *wn)
{
	struct private_alsa_data *pad = wn->private_data;
	snd_pcm_drain(pad->handle);
	snd_pcm_close(pad->handle);
	snd_config_update_free_global();
	free(pad);
}

/** the init function of the alsa writer */
void alsa_writer_init(struct writer *w)
{
	w->open = alsa_open;
	w->write = alsa_write;
	w->close = alsa_close;
	w->shutdown = NULL; /* nothing to do */
}
