/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file ortp_send.c para_server's ortp sender */

#include <ortp/ortp.h>

#include "server.cmdline.h"
#include "server.h"
#include "afs.h"
#include "send.h"
#include "list.h"
#include "ortp.h"
#include "string.h"

/** \cond convert in_addr to ascii */
#define TARGET_ADDR(oc) inet_ntoa((oc)->addr)
/** \endcond */

extern struct gengetopt_args_info conf;

/** describes one entry in the list of targets for the ortp sender */
struct ortp_target {
/** address info */
	struct in_addr addr;
/** whether the ortp sender is activated */
	int status;
/** the ortp timestamp increases by this amount */
	int chunk_ts;
/** the currently used timestamp for this target */
	int last_ts;
/** the position of this target in the list of targets */
	struct list_head node;
/** the UDP port */
	int port;
/** non-zero if at least one chunk has been sent to this target */
	int streaming;
/** the session pointer from libortp */
	RtpSession *session;
};

static struct list_head targets;
static struct sender *self;

static void ortp_delete_target(struct ortp_target *ot, const char *msg)
{
	PARA_NOTICE_LOG("deleting %s:%d (%s) from list\n", TARGET_ADDR(ot),
		ot->port, msg);
	if (ot->session) {
		rtp_session_destroy(ot->session);
		ot->session = NULL;
	}
	list_del(&ot->node);
	free(ot);
}

static void ortp_send_buf(char *buf, int len, long unsigned chunks_sent)
{
	struct ortp_target *ot, *tmp;
	int ret;

	list_for_each_entry_safe(ot, tmp, &targets, node) {
		int ts;
		if (!ot->session)
			continue;
		WRITE_CHUNK_TS(buf, ot->chunk_ts);
		ts = ot->chunk_ts * chunks_sent;
		ret = rtp_session_send_with_ts(ot->session, buf, len, ts);
		ot->last_ts = ts;
		if (ret < 0)
			ortp_delete_target(ot, "send error");
		if (ret != len +12)
			PARA_NOTICE_LOG("short write %d\n", ret);
	}
}

static void ortp_init_session(struct ortp_target *ot)
{
	RtpSession *s;
	int ret;

	PARA_NOTICE_LOG("sending to udp %s:%d\n", TARGET_ADDR(ot), ot->port);
	ot->session = rtp_session_new(RTP_SESSION_SENDONLY);
	if (!ot->session)
		return;
	s = ot->session;
	if (conf.ortp_jitter_compensation_arg) {
		rtp_session_enable_adaptive_jitter_compensation(ot->session, TRUE);
		rtp_session_set_jitter_compensation(ot->session,
			conf.ortp_jitter_compensation_arg);
	}
	/* always successful */
	rtp_session_set_send_payload_type(s, PAYLOAD_AUDIO_CONTINUOUS);
	ret = rtp_session_set_remote_addr(s, TARGET_ADDR(ot), ot->port);
	if (ret < 0) {
		rtp_session_destroy(ot->session);
		ot->session = NULL;
	}
}

/* called by afs */
static void ortp_shutdown_targets(void)
{
	char buf[ORTP_AUDIO_HEADER_LEN];
	struct ortp_target *ot, *tmp;

	WRITE_PACKET_TYPE(buf, ORTP_EOF);
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (!ot->session || !ot->streaming)
			continue;
		PARA_INFO_LOG("sending eof to ortp target %s:%d, ts = %d\n",
			TARGET_ADDR(ot), ot->port, ot->last_ts);
		rtp_session_send_with_ts(ot->session, buf,
			ORTP_AUDIO_HEADER_LEN, ot->last_ts);
		ot->streaming = 0;
		ot->chunk_ts = 0;
		rtp_session_reset(ot->session);
	}
}

static int need_extra_header(struct audio_format *af, long unsigned chunks_sent)
{
	/* FIXME: No need to compute this on every run */
	int mod = conf.ortp_header_interval_arg / (tv2ms(&af->chunk_tv) + 1);
	if (mod && (chunks_sent % mod))
		return 0;
	return 1;
}

static void ortp_send(struct audio_format *af, long unsigned current_chunk,
		long unsigned chunks_sent, const char *buf, size_t len)
{
	struct ortp_target *ot, *tmp;
	size_t sendbuf_len;
	int header_len = 0;
	int packet_type = ORTP_DATA, stream_type = af && af->get_header_info; /* header stream? */
	char *sendbuf, *header_buf = NULL;

	if (self->status != SENDER_ON)
		return;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (!ot->session) {
			ortp_init_session(ot);
			if (!ot->session)
				continue;
		}
		if (!ot->chunk_ts)
			ot->chunk_ts = rtp_session_time_to_ts(ot->session,
				tv2ms(&af->chunk_tv));
//		PARA_DEBUG_LOG("len: %d, ts: %lu, ts: %d\n",
//			len, ot->chunk_ts * chunks_sent, ot->chunk_ts);
		ot->streaming = 1;
	}
	if (list_empty(&targets))
		return;
	if (stream_type) {
		header_buf = af->get_header_info(&header_len);
		if (!need_extra_header(af, chunks_sent))
			header_len = 0;
	}
	sendbuf_len = ORTP_AUDIO_HEADER_LEN + header_len + len;
	sendbuf = para_malloc(sendbuf_len);
	if (!current_chunk)
		packet_type = ORTP_BOF;
	else if (header_len)
		packet_type = ORTP_HEADER;
	WRITE_PACKET_TYPE(sendbuf, packet_type);
	WRITE_CHUNK_TIME(sendbuf, af->chunk_tv.tv_usec);
	WRITE_STREAM_TYPE(sendbuf, stream_type);
	WRITE_HEADER_LEN(sendbuf, header_len);
	if (header_len)
		memcpy(sendbuf + ORTP_AUDIO_HEADER_LEN, header_buf,
			header_len);
	memcpy(sendbuf + ORTP_AUDIO_HEADER_LEN + header_len, buf, len);
	ortp_send_buf(sendbuf, sendbuf_len, chunks_sent);
	free(sendbuf);
}

static int ortp_com_on(__a_unused struct sender_command_data *scd)
{

	self->status = SENDER_ON;
	return 1;
}

static int ortp_com_off(__a_unused struct sender_command_data *scd)
{
	ortp_shutdown_targets();
	self->status = SENDER_OFF;
	return 1;
}

static int ortp_com_delete(struct sender_command_data *scd)
{
	char *a = para_strdup(inet_ntoa(scd->addr));
	struct ortp_target *ot, *tmp;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (scd->port != ot->port)
			continue;
		if (strcmp(TARGET_ADDR(ot), a))
			continue;
		ortp_delete_target(ot, "com_delete");
	}
	return 1;
}

static void ortp_add_target(int port, struct in_addr *addr)
{
	struct ortp_target *ot = para_calloc(sizeof(struct ortp_target));
	ot->port = port;
	ot->addr = *addr;
	PARA_INFO_LOG("adding to target list (%s:%d)\n",
		TARGET_ADDR(ot), ot->port);
	list_add(&ot->node, &targets);
}

static int ortp_com_add(struct sender_command_data *scd)
{
	int port = (scd->port > 0)? scd->port : conf.ortp_default_port_arg;
	ortp_add_target(port, &scd->addr);
	return 1;
}

static char *ortp_info(void)
{
	struct ortp_target *ot;
	char *ret, *tgts = NULL;

	list_for_each_entry(ot, &targets, node) {
		char *tmp = make_message("%s%s:%d ", tgts? tgts : "",
			TARGET_ADDR(ot), ot->port);
		free(tgts);
		tgts = tmp;
	}
	ret = make_message(
		"ortp status: %s\n"
		"ortp default port: udp %d\n"
		"ortp targets: %s\n",
		(self->status == SENDER_ON)? "on" : "off",
		conf.ortp_default_port_arg,
		tgts? tgts : "(none)"
	);
	free(tgts);
	return ret;
}

static void ortp_init_target_list(void)
{
	int i;

	INIT_LIST_HEAD(&targets);
	for (i = 0; i < conf.ortp_target_given; i++) {
		char *arg = para_strdup(conf.ortp_target_arg[i]);
		char *p = strchr(arg, ':');
		int port;
		struct in_addr addr;

		if (!p)
			goto err;
		*p = '\0';
		if (!inet_aton(arg, &addr))
			goto err;
		port = atoi(++p);
		if (port < 0 || port > 65535)
			port = conf.ortp_default_port_arg;
		ortp_add_target(port, &addr);
		goto success;
err:
		PARA_CRIT_LOG("syntax error for ortp_target option "
			"#%d, ignoring\n", i);
success:
		free(arg);
		continue;
	}
}

static void ortp_pre_select(__a_unused struct audio_format *af,
	__a_unused  int *max_fileno, __a_unused fd_set *rfds,
	__a_unused fd_set *wfds)
{
	return;
}

static char *ortp_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {add|delete} IP port\n"
		"example: add 224.0.1.38 1500 (LAN-streaming)\n"
	);
}

/**
 * the init function of para_server's ortp sender
 *
 * \param s pointer to the http sender struct
 *
 * It initializes all function pointers of \a s and the list of ortp targets.
 */
void ortp_send_init(struct sender *s)
{
	ortp_init();
	INIT_LIST_HEAD(&targets);
	s->info = ortp_info;
	s->help = ortp_help;
	s->send = ortp_send;
	s->pre_select = ortp_pre_select;
	s->post_select = NULL;
	s->shutdown_clients = ortp_shutdown_targets;
	s->client_cmds[SENDER_ON] = ortp_com_on;
	s->client_cmds[SENDER_OFF] = ortp_com_off;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = ortp_com_add;
	s->client_cmds[SENDER_DELETE] = ortp_com_delete;
	self = s;
	s->status = SENDER_OFF;
	ortp_init_target_list();
	if (!conf.ortp_no_autostart_given)
		s->status = SENDER_ON;
	PARA_DEBUG_LOG("%s", "ortp sender init complete\n");
}
