/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include "string.h"
#include "write.cmdline.h"
#include "write.h"
#include "write_common.h"
#include "fd.h"

#include <sys/time.h> /* gettimeofday */

#include "error.h"

#define WAV_HEADER_LEN 44

static char *audiobuf;
static struct timeval *start_time;
struct gengetopt_args_info conf;

INIT_WRITE_ERRLISTS;

void para_log(int ll, const char* fmt,...)
{
	va_list argp;

	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

/**
 * read WAV_HEADER_LEN bytes from stdin to audio buffer
 *
 * \return -E_READ_HDR on errors and on eof before WAV_HEADER_LEN could be
 * read. A positive return value indicates success.
 */
static int read_wav_header(void)
{
	ssize_t ret, count = 0;

	while (count < WAV_HEADER_LEN) {
		ret = read(STDIN_FILENO, audiobuf + count, WAV_HEADER_LEN - count);
		if (ret <= 0)
			return -E_READ_HDR;
		count += ret;
	}
	return 1;
}

/**
 * check if current time is later than start_time
 * \param diff pointer to write remaining time to
 *
 * If start_time was not given, or current time is later than given
 * start_time, return 0. Otherwise, return 1 and write the time
 * difference between current time and start_time to diff. diff may be
 * NULL.
 *
 */
static int start_time_in_future(struct timeval *diff)
{
	struct timeval now;

	if (!conf.start_time_given)
		return 0;
	gettimeofday(&now, NULL);
	return tv_diff(start_time, &now, diff) > 0? 1 : 0;
}

/**
 * sleep until time given at command line
 *
 * This is called if the initial buffer is filled. It returns
 * immediately if no start_time was given at the command line
 * or if the given start time is in the past.
 *
 */
static void do_initial_delay(struct timeval *delay)
{
	do
		para_select(1, NULL, NULL, delay);
	while (start_time_in_future(delay));
}

static int read_stdin(char *buf, size_t bytes_to_load, size_t *loaded)
{
	ssize_t ret;

	while (*loaded < bytes_to_load) {
		ret = read(STDIN_FILENO, buf + *loaded, bytes_to_load - *loaded);
		if (ret <= 0) {
			if (ret < 0)
				ret = -E_READ_STDIN;
			return ret;
		}
		*loaded += ret;
	}
	return 1;
}
/**
 * play raw pcm data
 * \param loaded number of bytes already loaded
 *
 * If start_time was given, prebuffer data until buffer is full or
 * start_time is reached. In any case, do not start playing before
 * start_time.
 *
 * \return positive on success, negative on errors.
 */
static int pcm_write(struct writer_node_group *wng, size_t loaded)
{
	size_t bufsize, prebuf_size, bytes_to_load;
	struct timeval delay;
	int ret, not_yet_started = 1;

	ret = wng_open(wng);
	if (ret < 0)
		goto out;
	PARA_INFO_LOG("max chunk_bytes: %zd\n", wng->max_chunk_bytes);
	bufsize = (conf.bufsize_arg * 1024 / wng->max_chunk_bytes)
		* wng->max_chunk_bytes;
	audiobuf = para_realloc(audiobuf, bufsize);
	prebuf_size = conf.prebuffer_arg * bufsize / 100;
	bytes_to_load =  PARA_MAX(prebuf_size, wng->max_chunk_bytes);
	ret = read_stdin(audiobuf, bytes_to_load, &loaded);
	if (ret <= 0 || loaded < bytes_to_load) {
		if (ret >= 0)
			ret = -E_PREMATURE_END;
		goto out;
	}
	if (not_yet_started && start_time && start_time_in_future(&delay))
		do_initial_delay(&delay);
	not_yet_started = 0;
again:
	ret = wng_write(wng, audiobuf, &loaded);
	if (ret <= 0)
		goto out;
	ret = -E_WRITE_OVERRUN;
	if (loaded >= bufsize)
		goto out;
	bytes_to_load = PARA_MIN(wng->max_chunk_bytes, bufsize);
	ret = read_stdin(audiobuf, bytes_to_load, &loaded);
	if (ret < 0)
		goto out;
	if (!ret)
		wng->eof = 1;
	goto again;
out:
	wng_close(wng);
	return ret;
}

static struct writer_node_group *check_args(void)
{
	int i, ret = -E_WRITE_SYNTAX;
	static struct timeval tv;
	struct writer_node_group *wng = NULL;

	if (conf.list_writers_given) {
		char *msg = NULL;
		FOR_EACH_WRITER(i) {
			char *tmp = make_message("%s%s%s",
				i? msg : "",
				i? " " : "",
				writer_names[i]);
			free(msg);
			msg = tmp;
		}
		fprintf(stderr, "%s\n", msg);
		free(msg);
		exit(EXIT_SUCCESS);
	}
	if (conf.prebuffer_arg < 0 || conf.prebuffer_arg > 100)
		goto out;
	if (conf.start_time_given) {
		long unsigned sec, usec;
		if (sscanf(conf.start_time_arg, "%lu:%lu",
				&sec, &usec) != 2)
			goto out;
		tv.tv_sec = sec;
		tv.tv_usec = usec;
		start_time = &tv;
	}
	if (!conf.writer_given) {
		wng = setup_default_wng();
		ret = 1;
		goto out;
	}
	wng = wng_new(conf.writer_given);
	for (i = 0; i < conf.writer_given; i++) {
		ret = check_writer_arg(conf.writer_arg[i]);
		if (ret < 0)
			goto out;
		wng->writer_nodes[i].writer = &writers[ret];
	}
	ret = 1;
out:
	if (ret > 0)
		return wng;
	free(wng);
	return NULL;
}

/**
 * test if audio buffer contains a valid wave header
 *
 * \return If not, return 0, otherwise, store number of channels and sample rate
 * in struct conf and return WAV_HEADER_LEN.
 */
static size_t check_wave(void)
{
	unsigned char *a = (unsigned char*)audiobuf;
	if (a[0] != 'R' || a[1] != 'I' || a[2] != 'F' || a[3] != 'F')
		return WAV_HEADER_LEN;
	conf.channels_arg = (unsigned) a[22];
	conf.sample_rate_arg = a[24] + (a[25] << 8) + (a[26] << 16) + (a[27] << 24);
	return 0;
}

int main(int argc, char *argv[])
{
	int ret = -E_WRITE_SYNTAX;
	struct writer_node_group *wng = NULL;

	cmdline_parser(argc, argv, &conf);
	wng = check_args();
	if (!wng)
		goto out;
	init_supported_writers();
	audiobuf = para_malloc(WAV_HEADER_LEN);
	ret = read_wav_header();
	if (ret < 0)
		goto out;
	ret = pcm_write(wng, check_wave());
out:
	wng_destroy(wng);
	free(audiobuf);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	return ret;
}
